#ifndef A65_H
#define A65_H

/*
		      6502 Cross-Assembler in Portable C

		   Copyright (c) 1986 William C. Colley, III

This header file contains the global constants and data type definitions for
all modules of the cross-assembler.
*/

#include <stdio.h>

/*  Boolean defines  */
#define TRUE (1)
#define FALSE (0)


/*  On 8-bit machines, the static type is as efficient as the register	*/
/*  type and far more efficient than the auto type.  On larger machines	*/
/*  such as the 8086 family, this is not necessarily the case.  To	*/
/*  let you experiment to see what generates the fastest, smallest code	*/
/*  for your machine, I have declared internal scratch variables in	*/
/*  functions "SCRATCH int", "SCRATCH unsigned", etc.  A SCRATCH	*/
/*  varible is made static below, but you might want to try register	*/
/*  instead.								*/

#define	SCRATCH		static

/*  A slow, but portable way of cracking an unsigned into its various	*/
/*  component parts:							*/

#define	clamp(u)	((u) &= 0xffff)
#define	high(u)		(((u) >> 8) & 0xff)
#define	low(u)		((u) & 0xff)
#define	word(u)		((u) & 0xffff)

/*  The longest source line the assembler can hold without exploding:	*/

#define	MAXLINE		255

/*  The maximum number of source files that can be open simultaneously:	*/

#define	FILES		4

/*  The fatal error messages generated by the assembler:		*/

#define	ASMOPEN		"Source File Did Not Open"
#define	ASMREAD		"Error Reading Source File"
#define EXPOPEN		"Export File Did Not Open"
#define	DSKFULL		"Disk or Directory Full"
#define	FLOFLOW		"File Stack Overflow"
#define	HEXOPEN		"Object File Did Not Open"
#define	IFOFLOW		"If Stack Overflow"
#define	LSTOPEN		"Listing File Did Not Open"
#define	NOASM		"No Source File Specified"
#define NOEXP		"No Export File Specified"
#define	SYMBOLS		"Too Many Symbols"

/*  The warning messages generated by the assembler:			*/

#define	BADOPT		"Illegal Option Ignored"
#define	NOHEX		"-o Option Ignored -- No File Name"
#define	NOLST		"-l Option Ignored -- No File Name"
#define	TWOASM		"Extra Source File Ignored"
#define TWOEXP		"Extra Export File Ignored"
#define	TWOHEX		"Extra Object File Ignored"
#define	TWOLST		"Extra Listing File Ignored"

/*  Line assembler (A65.C) constants:					*/

#define	BIGINST		3		/*  longest instruction length	*/
#define	IFDEPTH		16		/*  maximum IF nesting level	*/
#define	NOP		0xea		/*  processor's NOP opcode	*/
#define	ON		1		/*  assembly turned on		*/
#define	OFF		-1		/*  assembly turned off		*/

/*  Line assembler (A65.C) opcode attribute word flag masks:		*/

#define	PSEUDO		0x80	/*  is pseudo op			*/
#define	ISIF		0x40	/*  is IF, ELSE, or ENDIF		*/

/*  Line assembler (A65.C) pseudo-op opcode token values:		*/

typedef enum {
	ALIGN = 1,
	BASE,
	DATE,
	DB,
	DW,
	ELSE,
	END,
	ENDI,
	EQU,
	EXP,
	IF,
	INCB,
	INCL,
	MSG,
	ORG,
	PAGE,
	RMB,
	SET,
	TITL
} PSEUDO_OP;

/*  Line assembler (A65.C) machine opcode attribute values:		*/

#define	CPXY		0
#define	BITOP		1
#define	INHOP		2
#define	JUMP		3
#define	CALL		4
#define	LOGOP		5
#define	RELBR		6
#define	STXY		7
#define	TWOOP		8
#define	LDXY		9
#define	INCOP		10

/* Line assembler (a65.c) file struct */
typedef struct {
	FILE *fp;
	char filename[MAXLINE];
	int linenum;
} FILE_INFO;

/*  Opcode argument field parser (A65EVAL.C) attribute bit masks:	*/

#define	ARGA		040	/*  A register specified.		*/
#define	ARGX		020	/*  X register specified.		*/
#define	ARGY		010	/*  Y register specified.		*/
#define	ARGIND		004	/*  Argument field in parentheses.	*/
#define	ARGIMM		002	/*  Immediate addressing specified.	*/
#define	ARGNUM		001	/*  Argument value present.		*/

/*  Lexical analyzer (A65EVAL.C) token buffer and stream pointer:	*/

typedef struct {
    unsigned attr;
    unsigned valu;
    char sval[MAXLINE + 1];
} TOKEN;

/*  Lexical analyzer (A65EVAL.C) token attribute values:		*/

typedef enum {
	EOL = 0,		/* end of line */
	SEP,			/* field separator */
	OPR,			/* operator */
	STR,			/* character string */
	VAL,			/* value */
	IMM,			/* immediate designator */
	REG,			/* register designator */
	ABS,			/* force absolute addressing */
} LEX_ATTR;

/*  Lexical analyzer (A65EVAL.C) token attribute word flag masks:	*/

#define	BINARY		0x8000	/*  Operator:	is binary operator	*/
#define	UNARY		0x4000	/*		is unary operator	*/
#define	PREC		0x0f00	/*		precedence		*/

#define	FORWD		0x8000	/*  Value:	is forward referenced	*/
#define	SOFT		0x4000	/*		is redefinable		*/

#define	TYPE		0x000f	/*  All:	token type		*/

/*  Lexical analyzer (A65EVAL.C) operator token values (unlisted ones	*/
/*  use ASCII characters):						*/

typedef enum {
	AND,
	GE,
	HIGH,
	LE,
	LOW,
	MOD,
	NE,
	NOT,
	OR,
	SHR,
	SHL,
	XOR,
} LEX_OP;

/*  Lexical analyzer (A65EVAL.C) operator precedence values:		*/

#define	UOP1		0x0000	/*  unary +, unary -			*/
#define	MULT		0x0100	/*  *, /, MOD, SHL, SHR			*/
#define	ADDIT		0x0200	/*  binary +, binary -			*/
#define	RELAT		0x0300	/*  >, >=, =, <=, <, <>			*/
#define	UOP2		0x0400	/*  NOT					*/
#define	LOG1		0x0500	/*  AND					*/
#define	LOG2		0x0600	/*  OR, XOR				*/
#define	UOP3		0x0700	/*  HIGH, LOW				*/
#define	START2		0x0800	/*  beginning of subexpression		*/
#define	RPREN		0x0900	/*  )					*/
#define	LPREN		0x0a00	/*  (					*/
#define	ENDEX		0x0b00	/*  end of expression			*/
#define	START		0x0c00	/*  beginning of expression		*/

/*  Utility package (A65UTIL.C) symbol table routines:			*/

struct _symbol {
    unsigned attr;
    unsigned valu;
    struct _symbol *left, *right;
    char sname[1];
};

typedef struct _symbol SYMBOL;

#define	SYMCOLS		4

/*  Utility package (A65UTIL.C) opcode/operator table routines:		*/

typedef struct {
    unsigned attr;
    unsigned valu;
    char oname[6];
} OPCODE;

/*  Utility package (A65UTIL.C) hex file output routines:		*/

#define	HEXSIZE		8192

#endif
